# -*- coding: utf-8 -*-

"""
Base CLI commands for all objects.

Model-specific objects and argument parsers will be defined in subclasses or by
way of factory methods.
"""

from __future__ import unicode_literals
import datetime
import logging
import os
import sys
import textwrap

import pynsot
from . import client
from .util import get_result

from .vendor import click, netaddr, prettytable
from .vendor.slumber.exceptions import (HttpClientError, HttpServerError)


# Constants/Globals
if os.getenv('PYNSOT_DEBUG'):
    logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

# Make the --help option also have -h
CONTEXT_SETTINGS = {
    'help_option_names': ['-h', '--help'],
}

# Tuple of HTTP errors used for exception handling
HTTP_ERRORS = (HttpClientError, HttpServerError)

# Where to find the command plugins.
CMD_FOLDER = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'commands')
)


# Mapping of resource_names to natural_keys. This is primarily used by
# App.get_single_object() to map parameters to a single object.
NATURAL_KEYS = {
    'devices': ['hostname'],
    'networks': ['cidr'],
    'attributes': ['name', 'resource_name'],
    'interfaces': ['name', 'device'],
}

# Mapping of resource_names to what we want objects to look like when formatted
# to stdout for grep-friendliness. These are string interpolation style.
GREP_FORMATS = {
    'devices': '%(hostname)s',
    'networks': '%(network_address)s/%(prefix_length)s',
    'addresses': '%(network_address)s/%(prefix_length)s',  # Same as networks
    'assignments': '%(hostname)s:%(interface_name)s:%(address)s',
    'attributes': '%(resource_name)s:%(name)s',
    'circuits': '%(name)s',
    'interfaces': '%(device_hostname)s:%(name)s',
    'sites': '%(name)s',
}


__all__ = (
    'NsotCLI', 'App', 'app'
)


class NsotCLI(click.MultiCommand):
    """
    Base command object used to define object-specific command-line parsers.

    This will load command plugins from the "commands" folder.

    Plugins must be named "cmd_{foo}.py" and must have a top-level command
    named "cli".
    """
    def list_commands(self, ctx):
        """List all commands from python modules in plugin folder."""
        rv = []
        for filename in os.listdir(CMD_FOLDER):
            if filename.endswith('.py') and filename.startswith('cmd_'):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Import a command module and return it."""
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('pynsot.commands.cmd_' + name,
                             None, None, ['app'])
        except ImportError as err:
            print err
            return None
        return mod.cli  # Each cmd_ plugin defines top-level "cli" command


class App(object):
    """Context object for holding state data for the CLI app."""
    def __init__(self, ctx, client_args=None, verbose=False):
        if client_args is None:
            client_args = {}

        # Force api_version 1.0 for CLI util.
        client_args['extra_args'] = {'api_version': '1.0'}

        self.client_args = client_args
        self.ctx = ctx
        self.verbose = verbose
        self.resource_name = self.ctx.invoked_subcommand
        self.grep_name = self.resource_name
        self.site_id = None  # This is populated later.
        self.rebase_done = False  # So that we only rebase once.

    @property
    def api(self):
        """This way the API client is not created until called."""
        if not hasattr(self, '_api'):
            self._api = client.get_api_client(**self.client_args)
        return self._api

    @property
    def singular(self):
        """Return singular form of resource_name. (e.g. "sites" -> "site")"""
        resource_name = self.resource_name
        if resource_name.endswith('s'):
            resource_name = resource_name[:-1]
        return resource_name

    @property
    def resource(self):
        """
        Return an API resource method for calling endpoints.

        For example if ``resource_name`` is ``networks``, this is equivalent to
        calling ``self.api.networks``.
        """
        return self.api.get_resource(self.resource_name)

    @staticmethod
    def pretty_dict(data, delim='=', sep=', ', joiner='\n'):
        """
        Return a dict in k=v format. And also make it nice to look at.

        :param dict:
            A dict

        :param sep:
            Character used to separate items

        :param joiner:
            Character used to join items
        """
        log.debug('PRETTY DICT INCOMING DATA = %r', data)
        pretty = ''
        for key, val in data.iteritems():
            if isinstance(val, list):
                # Sort, add a newline and indent so that nested value items
                # look better.
                val = joiner.join(' ' + i for i in sorted(val))
                if val:
                    val = joiner + val  # Prefix it w/ newline for readability
            pretty += '%s%s%s%s' % (key, delim, val, sep)

        return pretty.rstrip(sep)  # Drop the trailing separator

    def format_message(self, obj_single, message=''):
        """
        Attempt to make messages human-readable.

        :param obj_single:
            Singular object name

        :param message:
            Dat message tho!
        """
        log.debug('FORMATTING MESSAGE: %r' % (message,))
        if 'UNIQUE constraint failed' in message:
            message = '%s object already exists.' % (obj_single.title(),)
        return message

    def format_timestamp(self, ts):
        """
        Take a UNIX timestamp and make it a datetime string.

        :param ts:
            UNIX timestamp
        """
        ts = int(ts)
        dt = datetime.datetime.fromtimestamp(ts)
        return str(dt)

    def handle_error(self, action, data, err):
        """
        Handle error API response.

        :param action:
            The action name

        :param data:
            Dict of arguments

        :param err:
            Exception object
        """
        resp = getattr(err, 'response', None)
        obj_single = self.singular
        extra = '\n'

        # If it's an API error, format it all pretty-like for human eyes.
        if resp is not None:
            body = resp.json()
            log.debug('API ERROR: %r' % (body,))

            msg = body['error']['message']
            msg = self.format_message(obj_single, msg)

            # Add the status code and reason to the output
            log.debug('ERROR MESSAGE = %r' % (msg,))
            if self.verbose or not msg:
                t_ = '%s %s'
                reason = resp.reason.upper()
                extra += t_ % (resp.status_code, reason)
            if not msg:
                msg = extra.strip()
        else:
            msg = str(err)

        if isinstance(msg, dict):
            msg = self.pretty_dict(msg, delim=': ', joiner='')

        # If we're being verbose, print some extra context.
        if self.verbose:
            t_ = ' trying to %s %s with args: %s'
            pretty_dict = self.pretty_dict(data)
            extra += t_ % (action, obj_single, pretty_dict)
            msg += extra

        # Colorize the failure text as red.
        self.ctx.exit(click.style('[FAILURE] ', fg='red') + msg)

    def handle_response(self, action, data, result):
        """
        Handle positive API response.

        :param action:
            The action name

        :param data:
            Dict of arguments

        :param result:
            Dict containing result
        """
        if isinstance(data, list):
            for item in data:
                self.handle_response(action, item, result)
            return None

        t_ = '%s %s!'
        if action.endswith('e'):
            action = action[:-1]  # "remove" -> "remov"
        action = action.title() + 'ed'  # "remove" -> "removed"
        msg = t_ % (action, self.singular)

        # Colorize the success text as green.
        click.echo(click.style('[SUCCESS] ', fg='green') + msg)

    def map_fields(self, fields, fields_map):
        """
        Map ``fields`` using ``fields_map`` for table display.

        :param fields:
            List of field names

        :param fields_map:
            Mapping of field names to translations
        """
        log.debug('MAP_FIELDS FIELDS = %r' % (fields,))
        log.debug('MAP_FIELDS FIELDS_MAP = %r' % (fields_map,))
        try:
            headers = [fields_map[f] for f in fields]
        except KeyError as err:
            msg = 'Could not map field %s when displaying results.' % (err,)
            self.ctx.exit(msg)
        log.debug('MAP_FIELDS HEADERS = %r' % (headers,))
        return headers

    def format_field(self, field, field_data):
        """
        Specially format a field.

        :param field:
            Field name

        :param field_data:
            Field data dict
        """
        # If it's a user field, only show the email
        if field == 'user':
            field_data = field_data['email']

        # Arbitrarily wrap description field at 20 chars
        if field == 'description':
            field_data = '\n'.join(textwrap.wrap(field_data, 20))

        # If the field is a dict, pretty_dict it!
        if isinstance(field_data, dict):
            # If this is an inner dict, prettify it, too.
            for sub_field in ('attributes', 'constraints'):
                if sub_field in field_data:
                    field_data[sub_field] = self.pretty_dict(
                        field_data[sub_field], sep='\n')

            # This is so that k=v looks better when printing a resource's
            # contents
            if field == 'resource':
                delim = ':'
            # Otherwise we just fallback to standard k=v
            else:
                delim = '='
            field_data = self.pretty_dict(field_data, delim=delim, sep='\n')

        if isinstance(field_data, list):
            field_data = '\n'.join(field_data)

        # If it's the 'change_at' field, make it human-readable
        elif field == 'change_at':  # Timestamp
            field_data = self.format_timestamp(field_data)

        return field_data

    def format_object_for_grep(self, obj):
        """
        Formats an object... for grep.

        :param obj:
            Object
        """
        # If there's a "grep_name", use that, otherwise use resource_name
        grep_name = getattr(self, 'grep_name', self.resource_name)

        return GREP_FORMATS[grep_name] % obj

    def print_grep(self, objects):
        """
        Print a list of objects in a grep-friendly format.

        Attributes are displayed in key=value style, one per line.

        :param objects:
            List of objects
        """
        output = []
        for obj in objects:
            prefix = self.format_object_for_grep(obj)
            attrs = obj.get('attributes', {})
            keys = sorted(attrs)
            for k in keys:
                output.append('%s %s=%s' % (prefix, k, attrs[k]))

        click.echo('\n'.join(output))

    def print_by_natural_key(self, objects, delimiter='\n'):
        """
        Print a list of objects by their natural_key

        :param objects:
            List of objects

        :param delimiter:
            Character used to delimit objects
        """
        output = []
        for obj in objects:
            output.append(self.format_object_for_grep(obj))

        # Networks results must be specially sorted.
        if self.grep_name != 'networks':
            output = sorted(output)
        else:
            networks = sorted(
                netaddr.IPNetwork(d) for d in output
            )
            output = (str(n) for n in networks)

        click.echo(delimiter.join(output))

    def print_list(self, objects, display_fields):
        """
        Print a list of objects in a table format.

        :param objects:
            List of object dicts

        :param display_fields:
            Ordered list of 2-tuples of (field, display_name) used
            to translate field names for display
        """
        # Extract the field names and create a mapping used for translation
        fields = [f[0] for f in display_fields]  # Field names are 1st item
        fields_map = dict(display_fields)

        # Human-readable field headings as they will be displayed
        headers = self.map_fields(fields, fields_map)

        # Order the object key/val by the order in display fields
        table_data = []

        # We're doing all of this just so we can pretty print dicts as k=v
        for obj in objects:
            obj_list = []
            for field in fields:
                field_data = obj[field]

                # Attempt to format the field
                field_data = self.format_field(field, field_data)

                obj_list.append(field_data)
            table_data.append(obj_list)

        # Prepare the table object
        table = prettytable.PrettyTable(headers)

        # Display table in a frame
        table.vrules = prettytable.prettytable.FRAME

        # *or* Display table with row separators
        # table.hrules = prettytable.ALL
        # table.vrules = prettytable.NONE

        table.align = 'l'  # Left-align everything
        table.left_padding_width = 1

        # Add the table rows
        for row in table_data:
            table.add_row(row)

        # Only paginate if table is longer than terminal.
        _, t_height, = click.get_terminal_size()
        if len(table_data) > t_height:
            click.echo_via_pager(table)
        else:
            click.echo(table)

    def rebase(self, data):
        """
        If this is not a Site object, then rebase the API URL.

        :param data:
            Dict of query arguments
        """
        # Don't rebase again if we've already rebased.
        if self.rebase_done:
            return None

        # Handle bulk queries
        if isinstance(data, list):
            data = data[0]

        # Prefer site_id from args.
        site_id = data.pop('site_id', None)

        # Default to client's default_site provided in user's config or at CLI.
        if site_id is None and self.resource_name != 'sites':
            site_id = self.api.default_site

        log.debug('rebase: Got site_id: %s' % site_id)
        if site_id is not None:
            log.debug('rebase: Site_id found; rebasing API URL!')
            self.site_id = site_id
            self.api._store['base_url'] += '/sites/%s' % site_id

        # Mark rebase as done.
        self.rebase_done = True

    def add(self, data):
        """POST"""
        action = 'add'
        log.debug('adding %s' % data)
        self.rebase(data)

        try:
            result = self.resource.post(data)
        except HTTP_ERRORS as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)

    def get_single_object(self, data, resource=None, natural_keys=None):
        """
        Get a single object based on the natural key for this resource.

        :param data:
            Dict of query parameters

        :param resource:
            (Optional) API resource object

        :param natural_keys:
            (Optional) List of natural keys. If not provided, they are derived
            automatically from the resource.

        :returns:
            None
        """
        # Should be a list of natural keys for the resource
        if natural_keys is None:
            natural_keys = NATURAL_KEYS.get(self.resource_name, [])

        # Construct our params to retrieve a single object.
        params = {'limit': 1}
        for natural_key in natural_keys:
            natural_value = data.get(natural_key)

            # Only filter if the key was found. This is mostly for interfaces
            # where (device_id,) or (device_id, name) are valid lookups.
            if natural_value is not None:
                params[natural_key] = natural_value

        # Rebase before we toilet face.
        if resource is None:
            resource = self.resource

        # Get the results
        try:
            r = resource.get(**params)
        except HTTP_ERRORS:
            return None

        # Assert only 1 matching result.
        try:
            total = r['count']
        except KeyError:
            return r  # Single result such as from /api/networks/:id/parent/

        if total > 1:
            log.debug('get_single_object: More than one object found!')
            return None

        # Return a single result.
        r = get_result(r)
        try:
            return r[0]
        except IndexError:
            return None

    def detail(self, data, resource):
        """
        GET a detail endpoint and return the results.

        This method is currently only called for Network.next_address() and
        Network.next_network(), so its use-case is very specific.

        :param data:
            Dict of query parameters

        :param resource:
            A prepared API resource object
        """
        try:
            return resource.get(**data)
        except HTTP_ERRORS as err:
            self.handle_error('detail', data, err)

    def set_query(self, data, resource=None):
        """
        Get objects based on a set query for this resource.

        :param data:
            Dict of query parameters

        :param resource:
            (Optional) API resource object
        """

        self.rebase(data)

        if resource is None:
            resource = self.resource

        try:
            result = resource.query.get(**data)
        except HTTP_ERRORS as err:
            self.handle_error('list', data, err)

        return get_result(result)

    def natural_keys_by_query(self, data, delimited=False):
        """
        Run a set query and return the natural keys of the results.

        :param data:
            Dict of query parameters

        :param delimited:
            Whether to display the results as comma or newline delimited
        """
        objects = self.set_query(data)
        delimiter = ',' if delimited else '\n'
        self.print_by_natural_key(objects, delimiter)

    def list(self, data, display_fields=None, resource=None,
             verbose_fields=None):
        """
        GET objects and display them to stdout.

        :param data:
            Dict of query parameters

        :param display_fields:
            Mapping of field_names to list display

        :param resource:
            (Optional) API resource object

        :param verbose_fields:
            (Optional) Mapping of field_names to verbose detail display. If not
            provided, ``display_fields`` is used.
        """
        action = 'list'
        log.debug('listing %s' % data)
        obj_id = data.get('id')  # If obj_id, it's a single object

        grep = data.pop('grep', False)
        by_natural_key = data.pop('natural_key', False)

        # If a resource object is provided, call it instead, and only rebase if
        # we haven't provided our own resource.
        if resource is None:
            self.rebase(data)  # Rebase first
            resource = self.resource

        obj = None
        try:
            # Try getting a single object first
            if obj_id:
                log.debug('Retrieving by obj_id=%r' % obj_id)
                result = resource(obj_id).get()
                obj = get_result(result)

            # If we still don't have an object try param-based lookup.
            if obj is None:
                log.debug('Retrieving by single_object')
                obj = self.get_single_object(data, resource)

            # If obj is STILL None...
            if obj is not None:
                # Set display_fields to verbose
                display_fields = verbose_fields or display_fields
            else:
                log.debug('Retrieving by params=%r' % data)
                result = resource.get(**data)

        except HTTP_ERRORS as err:
            self.handle_error(action, data, err)

        else:
            objects = []

            # Turn a single object into a list
            if obj:
                objects = [obj]

            # Or just list all of them.
            elif result:
                objects = get_result(result)

            if objects:
                if grep:
                    self.print_grep(objects)
                elif by_natural_key:
                    self.print_by_natural_key(objects)
                else:
                    self.print_list(objects, display_fields)
            else:
                pretty_dict = self.pretty_dict(data)
                t_ = 'No %s found matching args: %s!'
                msg = t_ % (self.singular, pretty_dict)
                click.echo(msg)

    def remove(self, **data):
        """DELETE"""
        action = 'remove'
        obj_id = data['id']
        log.debug('removing %s' % obj_id)
        self.rebase(data)

        try:
            result = self.resource(obj_id).delete()
        except HTTP_ERRORS as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)

    def process_attributes(self, attrs, action, multi=False):
        """
        Return a merged dict of CLI attribute actions w/ existing attributes.

        Turn attribute/value pairs into actionable attributes that the API can
        consume.

        :param attrs:
            Dictionary of outgoing attributes to merge with

        :param action:
            The action to perform against the outgoing attributes

        :param multi:
            Whether to treat the incoming attributes as list types
        """
        log.debug('PROCESS_ATTRIBUTES [IN]: %r', attrs)

        log.debug('ATTRIBUTES = %r', self.ctx._attributes)
        log.debug('ATTR_ACTION = %r', action)
        log.debug('      MULTI = %r', multi)

        multi_replaced = False
        for key, val in self.ctx._attributes:
            # List-type handler
            if multi:
                if action == 'add':
                    log.debug('[multi] Adding %r to %r' % (val, key))
                    if key not in attrs:
                        attrs[key] = [val]
                    elif val not in attrs[key]:
                        attrs[key].append(val)
                elif action == 'replace':
                    log.debug('[multi] Replacing %r with [%r]' % (key, val))
                    if key in attrs and not multi_replaced:
                        attrs[key] = []
                        multi_replaced = True

                    if val not in attrs[key]:
                        attrs[key].append(val)
                elif action == 'delete':
                    log.debug('[multi] Removing %r from %r' % (val, key))
                    # Remove this item from the list
                    if val in attrs[key]:
                        attrs[key].remove(val)

                    # If the value wasn't provided, or if the list is now
                    # empty, remove it.
                    if not val or not attrs[key]:
                        attrs.pop(key)

            # Default string-based handler
            else:
                if action in ('add', 'replace'):
                    log.debug(
                        '[single] Adding/replacing %r to %r' % (val, key)
                    )
                    attrs[key] = val
                elif action == 'delete':
                    log.debug('[single] Removing %r from %r' % (val, key))
                    attrs.pop(key, None)

        log.debug('PROCESS_ATTRIBUTES [OUT]: %r', attrs)
        return attrs

    def update(self, data):
        """PUT"""

        action = 'update'
        obj_id = data.pop('id')
        attr_action = data.pop('attr_action', None)
        multi = data.get('multi', False)
        log.debug('updating %s' % data)
        self.rebase(data)

        # Get the original object by id first so that we can keep any existing
        # values without resetting them.
        try:
            if obj_id:
                log.debug('Retrieving by obj_id=%r' % obj_id)
                result = self.resource(obj_id).get()
                obj = get_result(result)
            else:
                obj = self.get_single_object(data)
        except HTTP_ERRORS as err:
            self.handle_error(action, data, err)
        else:
            # FIXME(jathan) This error case needs work. It needs to be as
            # descriptive as when a lookup by id fails.
            if obj is None:
                self.handle_error(
                    action, data,
                    'Update failed. Try again with --verbose for more info.'
                )

            obj_id = obj.pop('id')  # We don't want id when doing a PUT
            payload = obj

        log.debug('PAYLOAD [in]: %r', payload)

        # Update the payload from the CLI params if the value isn't null.
        for key, val in data.iteritems():
            # If we're updating attributes, reconcile with existing attributes
            if key == 'attributes':
                attrs = payload['attributes']
                payload['attributes'] = self.process_attributes(
                    attrs, attr_action, multi=multi
                )

            # Otherwise, if the value was provided, replace it outright
            elif val is not None:
                payload[key] = val

        log.debug('PAYLOAD [out]: %r', payload)
        # And now we call PUT
        try:
            result = self.resource(obj_id).put(payload)
            log.debug('RESULT [out]: %r', result)
        except HTTP_ERRORS as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)


@click.command(cls=NsotCLI, context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', is_flag=True, help='Toggle verbosity.')
@click.version_option(version=pynsot.__version__)
@click.pass_context
def app(ctx, verbose):
    """
    Network Source of Truth (NSoT) command-line utility.

    For detailed documentation, please visit https://nsot.readthedocs.io
    """
    # This is the "app" object attached to all contexts.
    ctx.obj = App(ctx=ctx, verbose=verbose)

    # Store the invoked_subcommand (e.g. 'networks') name as
    # parent_resource_name so that descendent sub-commands can reference where
    # they came from, such as when calling callbacks.list_subcommand()
    ctx.obj.parent_resource_name = ctx.invoked_subcommand


if __name__ == '__main__':
    app()



# -*- coding: utf-8 -*-

"""
Test Circuits in the CLI app.
"""

from __future__ import absolute_import, unicode_literals
import logging

from tests.fixtures import (attribute, attributes, client, config, device,
                            interface, network, runner, site, site_client)
from tests.fixtures.circuits import (circuit, circuit_attributes, device_a,
                                     device_z, interface_a, interface_z)
from tests.util import assert_output, assert_outputs


log = logging.getLogger(__name__)


def test_circuits_add(runner, interface_a, interface_z):
    """ Test adding a normal circuit """

    with runner.isolated_filesystem():
        # Add a circuit with interfaces on each end
        result = runner.run(
            'circuits add -A {0} -Z {1} -n add_test1'.format(
                interface_a['id'],
                interface_z['id']
            )
        )
        assert_output(result, ['Added circuit!'])

        # Verify the circuit was created by listing
        result = runner.run('circuits list')
        assert_output(result, ['add_test1'])


def test_circuits_add_single_sided(runner, interface_a):
    """ Add a circuit with no remote end """

    with runner.isolated_filesystem():
        result = runner.run(
            'circuits add -A {0} -n add_test2'.format(interface_a['id'])
        )

        assert_output(result, ['Added circuit!'])

        result = runner.run('circuits list')
        assert_output(result, ['add_test2'])


def test_circuits_add_intf_reuse(runner, interface_a):
    """
    Try creating two circuits with the same interface, which should fail
    """

    with runner.isolated_filesystem():
        cmd = 'circuits add -A {0} -n {1}'

        result = runner.run(cmd.format(interface_a['id'], 'circuit1'))
        assert result.exit_code == 0

        result = runner.run(cmd.format(interface_a['id'], 'bad_circuit'))
        assert result.exit_code != 0
        assert 'endpoint_a:  This field must be unique' in result.output


def test_circuits_add_dupe_name(runner, interface_a, interface_z):
    """
    Try creating two circuits with the same name, which should fail
    """

    with runner.isolated_filesystem():
        cmd = 'circuits add -A {0} -n foo'

        result = runner.run(cmd.format(interface_a['id']))
        assert result.exit_code == 0

        result = runner.run(cmd.format(interface_z['id']))
        assert result.exit_code != 0
        assert 'circuit with this name already exists' in result.output


def test_circuits_list(runner, circuit):
    """ Make sure we can list out a circuit """

    circuit_name = 'test_circuit'

    with runner.isolated_filesystem():
        result = runner.run('circuits list')
        assert_output(result, [circuit_name])

        result = runner.run('circuits list -i {}'.format(circuit_name))
        assert_output(result, [circuit_name])


def test_circuits_list_nonexistant(runner):
    """ Listing a non-existant circuit should fail """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -i nopenopenope')

        assert result.exit_code != 0
        assert 'No such Circuit found' in result.output


def test_circuits_list_natural_key_output(runner, circuit):
    """ Natural key output should just list the circuit names """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -N')

        assert result.exit_code == 0
        assert result.output == "test_circuit\n"


def test_circuits_list_grep_output(runner, circuit):
    """ grep output should list circuit names with all the attributes """

    expected_output = (
        "test_circuit owner=alice\n"
        "test_circuit vendor=lasers go pew pew\n"
    )

    with runner.isolated_filesystem():
        result = runner.run('circuits list -g')

        assert result.exit_code == 0
        assert result.output == expected_output


def test_circuits_list_addresses(runner, circuit, interface_a, interface_z):
    """ Test listing out a circuit's interface addresses """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -i {} addresses'.format(
            circuit['id']
        ))

        assert_outputs(
            result,
            [
                [interface_a['addresses'][0].split('/')[0]],
                [interface_z['addresses'][0].split('/')[0]]
            ]
        )


def test_circuits_list_devices(runner, circuit, device_a, device_z):
    """ Test listing out a circuit's devices """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -i {} devices'.format(
            circuit['id']
        ))

        assert_outputs(
            result,
            [
                [device_a['hostname']],
                [device_z['hostname']]
            ]
        )


def test_circuits_list_interfaces(runner, circuit, interface_a, interface_z):
    """ Test listing out a circuit's interfaces """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -i {} interfaces'.format(
            circuit['id']
        ))

        assert_outputs(
            result,
            [
                [interface_a['device_hostname'], interface_a['name']],
                [interface_z['device_hostname'], interface_z['name']]
            ]
        )


def test_circuits_subcommand_query(runner, circuit):
    """ Make sure we can run a subcommand given a unique set query """

    with runner.isolated_filesystem():
        result = runner.run('circuits list -q owner=alice interfaces')
        assert result.exit_code == 0


def test_circuits_remove(runner, circuit):
    """ Make sure we can remove an existing circuit """

    circuit_name = 'test_circuit'

    with runner.isolated_filesystem():
        result = runner.run('circuits remove -i {}'.format(circuit_name))
        assert result.exit_code == 0


def test_circuits_update_name(runner, circuit):
    """ Test update by changing the circuit name """

    old_name = 'test_circuit'
    new_name = 'awesome_circuit'

    with runner.isolated_filesystem():
        result = runner.run('circuits update -i {} -n {}'.format(
            old_name,
            new_name
        ))
        assert result.exit_code == 0

        # Make sure we can look up the circuit by its new name
        result = runner.run('circuits list -i {}'.format(new_name))
        assert_output(result, [new_name])

        # Make sure the old name doesn't exist
        result = runner.run('circuits list -i {}')
        assert result.exit_code != 0
        assert 'No such Circuit found' in result.output


def test_circuits_update_interface(runner, circuit, interface):
    """ Test updating a circuit's Z side interface """

    with runner.isolated_filesystem():
        result = runner.run('circuits update -i {0} -Z {1}'.format(
            circuit['name'], interface['id']
        ))
        assert_output(result, ['Updated circuit!'])



"""
This is a pseudo module that allows you to proxy imports. This lets you
import dependencies in the form `from .vendor import requests`. We provide
a hook through the environment variable _PYNSOT_PYTHONPATH to override
the path for `pynsot` dependencies, otherwise it uses sys.path by default.

There is a limitation with dependencies that use absolute imports in their
packages as they will not be imported under the pynsot.vendor namespace.
"""

import sys
import os

# os.path doesn't import cleanly with this system from dependencies.
# Manually import so that it's available in the sys.modules cache.
import os.path


def get_path():
    env = os.environ.get("_PYNSOT_PYTHONPATH")
    if env:
        return env.split(":")
    return sys.path


__path__ = get_path()



# -*- coding: utf-8 -*-

"""
Handle the read, write, and generation of the .pynsotrc config file.
"""

from __future__ import unicode_literals
from ConfigParser import RawConfigParser
import copy
import logging
import os

from .vendor import click
from .vendor import rcfile
from . import constants


__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015-2016 Dropbox, Inc.'


# Logging object
log = logging.getLogger(__name__)


__all__ = (
    'DotfileError', 'Dotfile',
)


class DotfileError(Exception):
    """Raised when something with the dotfile fails."""


class Dotfile(object):
    """Create, read, and write a dotfile."""
    def __init__(self, filepath=constants.DOTFILE_PATH, **kwargs):
        self.filepath = filepath

    def read(self, **kwargs):
        """
        Read ``~/.pynsotrc`` and return it as a dict.
        """
        config = rcfile.rcfile(
            constants.SECTION_NAME, args={'config': self.filepath}
        )
        config.pop('config', None)  # We don't need this field in here.

        # If there is *no* config data so far and...
        if not config and not os.path.exists(self.filepath):
            p = '%s not found; would you like to create it?' % (self.filepath,)
            if click.confirm(p, default=True, abort=True):
                config_data = self.get_config_data(**kwargs)
                self.write(config_data)  # Write config to disk
                config = config_data  # Return the contents

        self.config = config

        # If we have configuration values, validate the permissions and
        # presence of fields in the dotfile.
        auth_method = config.get('auth_method')
        if auth_method == 'auth_token':
            self.validate_perms()
            required_fields = self.get_required_fields(auth_method)
            self.validate_fields(config, required_fields)

        return config

    def validate_perms(self):
        """Make sure dotfile ownership and permissions are correct."""
        if not os.path.exists(self.filepath):
            return None

        # Ownership
        s = os.stat(self.filepath)
        if s.st_uid != os.getuid():
            raise DotfileError(
                '%s: %s must be owned by you' % (
                    constants.DOTFILE_NAME, self.filepath
                )
            )

        # Permissions
        self.enforce_perms()

    def validate_fields(self, field_names, required_fields):
        """
        Make sure all the fields are set.

        :param field_names:
            List of field names to validate

        :param required_fields:
            List of required field names to check against
        """
        for rname in sorted(required_fields):
            if rname not in field_names:
                msg = '%s: Missing required field: %s' % (self.filepath, rname)
                raise DotfileError(msg)

    def enforce_perms(self, perms=constants.DOTFILE_PERMS):
        """
        Enforce permissions on the dotfile.

        :param perms:
            Octal number representing permissions to enforce
        """
        log.debug('Enforcing permissions %o on %s' % (perms, self.filepath))
        os.chmod(self.filepath, perms)

    def write(self, config_data, filepath=None):
        """
        Create a dotfile from keyword arguments.

        :param config_data:
            Dict of config settings

        :param filepath:
            (Optional) Path to write
        """
        if filepath is None:
            filepath = self.filepath
        config = RawConfigParser()
        section = constants.SECTION_NAME
        config.add_section(section)

        # Set the config settings
        for key, val in config_data.iteritems():
            config.set(section, key, val)

        with open(filepath, 'wb') as dotfile:
            config.write(dotfile)

        self.enforce_perms()
        log.debug('wrote %s' % filepath)

    @classmethod
    def get_required_fields(cls, auth_method, required_fields=None):
        """
        Return union of all required fields for ``auth_method``.

        :param auth_method:
            Authentication method

        :param required_fields:
            Mapping of required field names to default values
        """
        if required_fields is None:
            required_fields = copy.deepcopy(constants.REQUIRED_FIELDS)

        from .client import AUTH_CLIENTS  # To avoid circular import

        # Fields that do not have default values, which are specific to an
        # auth_method.
        auth_fields = AUTH_CLIENTS[auth_method].required_arguments
        non_specific_fields = dict.fromkeys(auth_fields)
        required_fields.update(non_specific_fields)

        # Fields that MAY have default values, which are specific to an
        # auth_method, if any.
        specific_fields = constants.SPECIFIC_FIELDS.get(auth_method, {})
        required_fields.update(specific_fields)

        return required_fields

    @classmethod
    def get_config_data(cls, required_fields=None, optional_fields=None,
                        **kwargs):
        """
        Enumerate required fields and prompt for ones that weren't provided if
        they don't have default values.

        :param required_fields:
            Mapping of required field names to default values

        :param optional_fields:
            Mapping of optional field names to default values

        :param kwargs:
            Dict of prepared config settings

        :returns:
            Dict of config data
        """
        if required_fields is None:
            required_fields = constants.REQUIRED_FIELDS

        if optional_fields is None:
            optional_fields = constants.OPTIONAL_FIELDS

        config_data = {}

        # Base required fields
        cls.process_fields(config_data, required_fields, **kwargs)

        # Auth-related fields
        auth_method = config_data['auth_method']
        auth_fields = cls.get_required_fields(auth_method)
        cls.process_fields(config_data, auth_fields, **kwargs)

        # Optional fields
        cls.process_fields(
            config_data, optional_fields, optional=True, **kwargs
        )

        return config_data

    @staticmethod
    def process_fields(config_data, field_items, optional=False, **kwargs):
        """
        Enumerate fields to update ``config_data``.

        Any fields not already in ``config_data`` will be prompted for a value
        from ``field_items``.

        :param config_data:
            Dict of config data

        :param field_items:
            Dict of desired fields to be populated

        :param optional:
            Whether to consider values for ``field_items`` as optional

        :param kwargs:
            Keyword arguments of prepared values
        """
        for field, default_value in field_items.iteritems():
            prompt = 'Please enter %s' % (field,)

            # If it's already in the config data, move on.
            if field in config_data:
                continue

            # If the field was not provided
            if field not in kwargs:
                # If it does not have a default value prompt for it
                if default_value is None:
                    if not optional:
                        value = click.prompt(prompt, type=str)
                    else:
                        prompt += ' (optional)'
                        value = click.prompt(
                            prompt, default='', type=str, show_default=False
                        )

                # If the default_value is a string, prompt for it, but present
                # it as a default.
                elif isinstance(default_value, basestring):
                    value = click.prompt(
                        prompt, type=str, default=default_value
                    )

                # If it's a list of options, present the choices.
                elif isinstance(default_value, list):
                    choices = ', '.join(default_value)
                    prompt = 'Please choose %s [%s]' % (field, choices)
                    while True:
                        value = click.prompt(prompt, type=str)
                        if value not in default_value:
                            click.echo('Invalid choice: %s' % value)
                            continue
                        break
                else:
                    raise RuntimeError(
                        'Unexpected error condition when processing config '
                        'fields.'
                    )

            # Or, use the value provided in kwargs
            elif field in kwargs:
                value = kwargs[field]

            # Otherwise, use the default value
            else:
                value = default_value

            # If a value wasn't set, don't save it.
            if value == '':
                continue

            config_data[field] = value

        return config_data



# -*- coding: utf-8 -*-

"""
Test the API client.
"""

from __future__ import unicode_literals
import logging
import pytest

from pynsot.util import get_result
from .fixtures import config, client


__all__ = ('client', 'config', 'pytest')


log = logging.getLogger(__name__)


def test_authentication(client):
    """Test manual client authentication."""
    auth = {
        'email': client.config['email'],
        'secret_key': client.config['secret_key']
    }
    # Good
    resp = client.authenticate.post(auth)
    result = get_result(resp)
    assert 'auth_token' in result


def test_sites(client):
    """Test working with sites using the client."""
    site = client.sites.post({'name': 'Foo'})
    assert client.sites.get() == [site]
    assert client.sites(site['id']).get() == site



# -*- coding: utf-8 -*-

"""
Test the utils lib.
"""

import pytest  # noqa

from pynsot.util import slugify, validate_cidr


def test_validate_cidr():
    """Test ``validate_cidr()``."""
    # IPv4
    assert validate_cidr('0.0.0.0/0')
    assert validate_cidr('1.2.3.4/32')

    # IPv6
    assert validate_cidr('::/0')
    assert validate_cidr('fe8::/10')

    # Bad
    assert not validate_cidr('bogus')
    assert not validate_cidr(None)
    assert not validate_cidr(object())
    assert not validate_cidr({})
    assert not validate_cidr([])


def test_slugify():
    cases = [
        ('/', '_'),
        ('my cool string', 'my cool string'),
        ('Ethernet1/2', 'Ethernet1_2'),
        (
            'foo-bar1:xe-0/0/0.0_foo-bar2:xe-0/0/0.0',
            'foo-bar1:xe-0_0_0.0_foo-bar2:xe-0_0_0.0'
        ),
    ]

    for case, expected in cases:
        assert slugify(case) == expected



# -*- coding: utf-8 -*-

"""
Sub-command for Circuits
"""

from __future__ import absolute_import, unicode_literals
import logging

from pynsot.util import slugify
from pynsot.vendor import click
from . import callbacks
from .cmd_networks import DISPLAY_FIELDS as NETWORK_DISPLAY_FIELDS
from .cmd_interfaces import DISPLAY_FIELDS as INTERFACE_DISPLAY_FIELDS
from .cmd_devices import DISPLAY_FIELDS as DEVICE_DISPLAY_FIELDS


# Logger
log = logging.getLogger(__name__)

# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('name', 'Name'),
    ('endpoint_a', 'Endpoint A'),
    ('endpoint_z', 'Endpoint Z'),
    ('attributes', 'Attributes'),
)

# Fields to display when viewing a single record.
VERBOSE_FIELDS = DISPLAY_FIELDS


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Circuit objects.

    A Circuit resource represents a connection between two interfaces on a
    device.
    """


# Add
# TODO(npegg): support specifying interfaces using the natural key instead of
# by ID when the API supports that
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Circuit (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-A',
    '--endpoint-a',
    metavar='INTERFACE_ID',
    required=True,
    type=int,
    help='Unique ID of the interface of the A side of the Circuit',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Circuit.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Circuit is under. [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '-Z',
    '--endpoint-z',
    metavar='INTERFACE_ID',
    type=int,
    help='Unique ID of the interface on the Z side of the Circuit',
)
@click.pass_context
def add(ctx, attributes, endpoint_a, name, site_id, endpoint_z):
    """
    Add a new Circuit.

    You must provide the A side of the circuit using the -A/--endpoint-a
    option. The Z side is recommended but may be left blank, such as in cases
    where it is not an Interface that is tracked by NSoT (like a provider's
    interface).

    The name (-n/--name) is optional. If it is not specified, it will be
    generated for you in the form of:
    {device_a}:{interface_a}_{device_z}:{interface_z}

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.
    """

    data = ctx.params

    # Remove empty values to facilitate default assignment
    if name is None:
        data.pop('name')
    if endpoint_z is None:
        data.pop('endpoint_z')

    ctx.obj.add(data)


# List
@cli.group(invoke_without_command=True)
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='Filter Circuits by matching attributes (format: key=value).',
    multiple=True,
)
@click.option(
    '-A',
    '--endpoint-a',
    metavar='INTERFACE_ID',
    help='Filter to Circuits with endpoint_a interfaces that match this ID'
)
@click.option(
    '-g',
    '--grep',
    is_flag=True,
    help='Display list results in a grep-friendly format.',
    default=False,
    show_default=True,
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Circuit being retrieved.',
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    help='Limit result to N resources.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    help='Filter to Circuits matching this name.',
)
@click.option(
    '-N',
    '--natural-key',
    is_flag=True,
    help='Display list results by their natural key',
    default=False,
    show_default=True,
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-q',
    '--query',
    metavar='QUERY',
    help='Perform a set query using Attributes and output matching Circuits.'
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    help='Unique ID of the Site this Circuit is under.',
    callback=callbacks.process_site_id,
)
@click.option(
    '-Z',
    '--endpoint-z',
    metavar='INTERFACE_ID',
    help='Filter to Circuits with endpoint_z interfaces that match this ID'
)
@click.pass_context
def list(ctx, attributes, endpoint_a, endpoint_z, grep, id, limit, name,
         natural_key, offset, query, site_id):
    """
    List existing Circuits for a Site.

    You must either have a Site ID configured in your .pysnotrc file or specify
    one using the -s/--site-id option.

    When listing Circuits, all objects are displayed by default. You optionally
    may look up a single Circuit by ID or Name using the -i/--id option.

    You may limit the number of results using the -l/--limit option.
    """

    # If we get a name as an identifier, slugify it
    if ctx.params.get('id') and not ctx.params['id'].isdigit():
        ctx.params['id'] = slugify(ctx.params['id'])

    # Don't list interfaces if a subcommand is invoked
    if ctx.invoked_subcommand is None:
        ctx.obj.list(ctx.params, display_fields=DISPLAY_FIELDS)


@list.command()
@click.pass_context
def addresses(ctx, *args, **kwargs):
    """ Show addresses for the Interfaces on this Circuit. """

    callbacks.list_subcommand(
        ctx, display_fields=NETWORK_DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def devices(ctx, *args, **kwargs):
    """ Show Devices connected by this Circuit. """

    callbacks.list_subcommand(
        ctx, display_fields=DEVICE_DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def interfaces(ctx, *args, **kwargs):
    """ Show Interfaces connected by this Circuit. """

    callbacks.list_subcommand(
        ctx, display_fields=INTERFACE_DISPLAY_FIELDS, my_name=ctx.info_name
    )


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Circuit (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-A',
    '--endpoint-a',
    metavar='INTERFACE_ID',
    type=int,
    help='Unique ID of the interface of the A side of the Circuit',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Circuit being retrieved.',
    required=True,
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Circuit.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Circuit is under. [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '-Z',
    '--endpoint-z',
    metavar='INTERFACE_ID',
    type=int,
    help='Unique ID of the interface on the Z side of the Circuit',
)
@click.option(
    '--add-attributes',
    'attr_action',
    flag_value='add',
    default=True,
    help=(
        'Causes attributes to be added. This is the default and providing it '
        'will have no effect.'
    )
)
@click.option(
    '--delete-attributes',
    'attr_action',
    flag_value='delete',
    help=(
        'Causes attributes to be deleted instead of updated. If combined with'
        'with --multi the attribute will be deleted if either no value is '
        'provided, or if attribute no longer has an valid values.'
    ),
)
@click.option(
    '--replace-attributes',
    'attr_action',
    flag_value='replace',
    help=(
        'Causes attributes to be replaced instead of updated. If combined '
        'with --multi, the entire list will be replaced.'
    ),
)
@click.pass_context
def update(ctx, attributes, endpoint_a, id, name, site_id, endpoint_z,
           attr_action):
    """
    Update a Circuit.

    You must either have a Site ID configured in your .pysnotrc file or specify
    one using the -s/--site-id option.

    The -a/--attributes option may be provided multiple times, once for each
    key-value pair.

    When modifying attributes you have three actions to choose from:

    * Add (--add-attributes). This is the default behavior that will add
    attributes if they don't exist, or update them if they do.

    * Delete (--delete-attributes). This will cause attributes to be
    deleted. If combined with --multi the attribute will be deleted if
    either no value is provided, or if the attribute no longer contains a
    valid value.

    * Replace (--replace-attributes). This will cause attributes to
    replaced. If combined with --multi and multiple attributes of the same
    name are provided, only the last value provided will be used.
    """

    # If we get a name as an identifier, slugify it
    if ctx.params.get('id') and not ctx.params['id'].isdigit():
        ctx.params['id'] = slugify(ctx.params['id'])

    if not any([attributes, endpoint_a, name, endpoint_z]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)

    ctx.obj.update(ctx.params)


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Circuit being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Circuit is under.',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove a Circuit.

    You must either have a Site ID configured in your .pysnotrc file or specify
    one using the -s/--site-id option.

    When removing Circuits, all objects are displayed by default. You
    optionally may look up a single Circuit by ID or Name using the -i/--id
    option.
    """

    # If we get a name as an identifier, slugify it
    if ctx.params.get('id') and not ctx.params['id'].isdigit():
        ctx.params['id'] = slugify(ctx.params['id'])

    ctx.obj.remove(**ctx.params)



# -*- coding: utf-8 -*-

"""
Utilities for testing.
"""

from __future__ import unicode_literals
import collections
import contextlib
from itertools import islice
import logging
import os
import random
import shlex
import shutil
import socket
import struct
import tempfile

from pynsot.app import app
from pynsot import client
from pynsot import dotfile
from pynsot.vendor.click.testing import CliRunner as BaseCliRunner


log = logging.getLogger(__name__)

# Phony attributes to randomly generate for testing.
ATTRIBUTE_DATA = {
    'lifecycle': ['monitored', 'ignored'],
    'owner': ['jathan', 'gary', 'lisa', 'jimmy', 'bart', 'bob', 'alice'],
    'metro': ['lax', 'iad', 'sjc', 'tyo'],
    'foo': ['bar', 'baz', 'spam'],
}

# Used to store Attribute/value pairs
Attribute = collections.namedtuple('Attribute', 'name value')

# Hard-code the app name as 'nsot' to match the CLI util.
app.name = 'nsot'


class CliRunner(BaseCliRunner):
    """
    Subclass of CliRunner that also creates a .pynsotrc in the isolated
    filesystem.
    """
    def __init__(self, client_config, *args, **kwargs):
        self.client_config = client_config
        super(CliRunner, self).__init__(*args, **kwargs)

    @contextlib.contextmanager
    def isolated_filesystem(self):
        """
        A context manager that creates a temporary folder and changes
        the current working directory to it for isolated filesystem tests.
        """
        # If user config is found, back it up for duration of each test.
        config_path = os.path.expanduser('~/.pynsotrc')
        backup_path = config_path + '.orig'
        backed_up = False
        if os.path.exists(config_path):
            log.debug('Config found, backing up...')
            os.rename(config_path, backup_path)
            backed_up = True

        cwd = os.getcwd()
        t = tempfile.mkdtemp()
        os.chdir(t)
        rcfile = dotfile.Dotfile('.pynsotrc')
        rcfile.write(self.client_config)
        try:
            yield t
        finally:
            os.chdir(cwd)
            if backed_up:
                log.debug('Restoring original config.')
                os.rename(backup_path, config_path)  # Restore original
            try:
                shutil.rmtree(t)
            except (OSError, IOError):
                pass

    def run(self, command, **kwargs):
        """
        Shortcut to invoke to parse command and pass app along.

        :param command:
            Command args e.g. 'devices list'

        :param kwargs:
            Extra keyword arguments to pass to ``invoke()``
        """
        cmd_parts = shlex.split(command)
        result = self.invoke(app, cmd_parts, **kwargs)
        return result


def assert_output(result, expected, exit_code=0):
    """
    Assert that output matches the conditions.

    :param result:
        CliRunner result object

    :param expected:
        List/tuple of expected outputs

    :param exit_code:
        Expected exit code
    """
    if not isinstance(expected, (tuple, list)):
        raise TypeError('Expected must be a list or tuple')

    assert result.exit_code == exit_code
    output = result.output.splitlines()

    for line in output:
        # Assert that the expected items are found on the same line
        if not all((e in line) for e in expected):
            continue
        else:
            log.info('matched: %r', (expected,))
            break
    else:
        assert False


def assert_outputs(result, expected_list, exit_code=0):
    """
    Assert output over a list of of lists of expected outputs.

    :param result:
        CliRunner result object

    :param expected_list:
        List of lists/tuples of expected outputs

    :param exit_code:
        Expected exit code
    """
    for expected in expected_list:
        assert_output(result, expected, exit_code)


def rando():
    """Flip a coin."""
    return random.choice((True, False))


def take_n(n, iterable):
    "Return first n items of the iterable as a list"
    return list(islice(iterable, n))


def generate_hostnames(num_items=100):
    """
    Generate a random list of hostnames.

    :param num_items:
        Number of items to generate
    """

    for i in range(1, num_items + 1):
        yield 'host%s' % i


def generate_ipv4():
    """Generate a random IPv4 address."""
    return socket.inet_ntoa(struct.pack('>I', random.randint(1, 0xffffffff)))


def generate_ipv4list(num_items=100, include_hosts=False):
    """
    Generate a list of unique IPv4 addresses. This is a total hack.

    :param num_items:
        Number of items to generate

    :param include_hosts:
        Whether to include /32 addresses
    """
    ipset = set()
    # Keep iterating and hack together cidr prefixes if we detect empty
    # trailing octects. This is so lame that we'll mostly just end up with a
    # bunch of /24 networks.
    while len(ipset) < num_items:
        ip = generate_ipv4()
        if ip.startswith('0'):
            continue

        if ip.endswith('.0.0.0'):
            prefix = '/8'
        elif ip.endswith('.0.0'):
            prefix = '/16'
        elif ip.endswith('.0'):
            prefix = '/24'
        elif include_hosts:
            prefix = '/32'
        else:
            continue

        ip += prefix
        ipset.add(ip)

    return sorted(ipset)


def enumerate_attributes(resource_name, attributes=None):
    if attributes is None:
        attributes = ATTRIBUTE_DATA

    for name in attributes:
        yield {'name': name, 'resource_name': resource_name}


def generate_attributes(attributes=None, as_dict=True):
    """
    Randomly choose attributes and values for testing.

    :param attributes:
        Dictionary of attribute names and values

    :param as_dict:
        If set return a dict vs. list of Attribute objects
    """
    if attributes is None:
        attributes = ATTRIBUTE_DATA

    attrs = []
    for attr_name, attr_values in attributes.iteritems():
        if random.choice((True, False)):
            attr_value = random.choice(attr_values)
            attrs.append(Attribute(attr_name, attr_value))

    if as_dict:
        attrs = dict(attrs)

    return attrs


def generate_devices(num_items=100, with_attributes=True):
    """
    Return a list of dicts for Device creation.

    :param num_items:
        Number of items to generate

    :param with_attributes:
        Whether to include Attributes
    """
    hostnames = generate_hostnames(num_items)

    devices = []
    for hostname in hostnames:
        item = {'hostname': hostname}
        if with_attributes:
            item['attributes'] = generate_attributes()

        devices.append(item)

    return devices


def generate_interface(name, device_id=None, with_attributes=True,
                       addresses=None):
    """
    Return a list of dicts for Interface creation.

    :param device_id:
        The device_id for the Interface

    :param with_attributes:
        Whether to include Attributes

    :param addresses:
        List of addresses to assign to the Interface
    """
    speeds = (100, 1000, 10000, 40000)
    types = (6, 135, 136, 161)
    if addresses is None:
        addresses = []

    item = {
        'name': name,
        # 'device_id': device_id,
        'device': device_id,
        'speed': random.choice(speeds),
        'type': random.choice(types),
    }

    if with_attributes:
        item['attributes'] = generate_attributes()
    if addresses:
        item['addresses'] = addresses

    return item


def generate_interfaces(device_id=None, with_attributes=True,
                        address_pool=None):
    """
    Return a list of dicts for Interface creation.

    Will generate as many Interfaces as there are in the address_pool.

    :param device_id:
        The device_id for the Interface

    :param num_items:
        Number of items to generate

    :param with_attributes:
        Whether to include Attributes

    :param address_pool:
        Pool of addresses to assign
    """
    prefix = 'eth'

    if address_pool is None:
        address_pool = []

    interfaces = []
    for num, address in enumerate(address_pool):
        name = prefix + str(num)

        # Currently hard-coded to 1 address per interface.
        addresses = [address]

        item = generate_interface(
            name, device_id, with_attributes=with_attributes,
            addresses=addresses
        )
        interfaces.append(item)

    return interfaces


def generate_networks(num_items=100, with_attributes=True, include_hosts=False,
                      ipv4list=None):
    """
    Return a list of dicts for Network creation.

    :param num_items:
        Number of items to generate

    :param with_attributes:
        Whether to include Attributes
    """
    if ipv4list is None:
        ipv4list = generate_ipv4list(num_items, include_hosts=include_hosts)

    networks = []
    for cidr in ipv4list:
        item = {'cidr': str(cidr)}
        if with_attributes:
            item['attributes'] = generate_attributes()

        networks.append(item)

    return networks


def populate_sites(site_data):
    """Populate sites from fixture data."""
    api = client.Client('http://localhost:8990/api', email='jathan@localhost')

    results = []
    for d in site_data:
        try:
            result = api.sites.post(d)
        except Exception as err:
            print err, d['name']
        else:
            results.append(result)

    print 'Created', len(results), 'sites.'


def rando_set_action():
    """Return a random set theory query action."""
    return random.choice(['+', '-', ''])


def rando_set_query():
    """Return a random set theory query string."""
    action = rando_set_action()
    return ' '.join(
        action + '%s=%s' % (k, v) for k, v in generate_attributes().iteritems()
    )



# -*- coding: utf-8 -*-

"""
Callbacks used in handling command plugins.
"""

from __future__ import unicode_literals
import ast
import csv
import logging

from ..vendor import click


log = logging.getLogger(__name__)

# Objects that do not have attribtues
NO_ATTRIBUTES = ('attributes',)


def process_site_id(ctx, param, value):
    """
    Callback to attempt to get site_id from ``~/.pynsotrc`` if it's not
    provided using -s/--site-id.
    """
    log.debug('GOT DEFAULT_SITE: %s' % ctx.obj.api.default_site)
    log.debug('GOT PROVIDED SITE_ID: %s' % value)

    # Try to get site_id from the app config, or complain that it's not set.
    if value is None:
        default_site = ctx.obj.api.default_site
        if default_site is None:
            raise click.UsageError('Missing option "-s" / "--site-id".')
        value = default_site
    else:
        log.debug('Setting provided site_id as default_site.')
        ctx.obj.api.default_site = value
    return value


def process_constraints(data, constraint_fields):
    """
    Callback to move constrained fields from incoming data into a 'constraints'
    key.

    :param data:
        Incoming argument dict

    :param constraint_fields:
        Constrained fields to move into 'constraints' dict
    """
    # Always use a list so that we can handle bulk operations
    objects = data if isinstance(data, list) else [data]

    # Enforce that pattern is a string.
    if data['pattern'] is None:
        data['pattern'] = ''

    for obj in objects:
        constraints = {}
        for c_field in constraint_fields:
            try:
                c_value = obj.pop(c_field)
            except KeyError:
                continue
            else:
                # If the value is not set, translate it to False
                if c_value is None:
                    c_value = False
                constraints[c_field] = c_value
        obj['constraints'] = constraints
    return data


def transform_attributes(ctx, param, value):
    """Callback to turn attributes arguments into a dict."""
    attrs = {}
    log.debug('TRANSFORM_ATTRIBUTES [IN]: %r' % (value,))

    # If this is a simple string, make it a list.
    if isinstance(value, basestring):
        value = [value]

    # Flatten the attributes using a set to eliminate any duplicates.
    items = set(value)

    # Prepare the context object for storing attribute actions
    parent = ctx.find_root()
    if not hasattr(parent, '_attributes'):
        parent._attributes = []

    for attr in items:
        key, _, val = attr.partition('=')
        if not key:
            msg = 'Invalid attribute: %s; format should be key=value' % (attr,)
            raise click.UsageError(msg)

        # Cast integers to strings (fix #24)
        if isinstance(val, int):
            val = str(val)

        log.debug(' name = %r', key)
        log.debug('value = %r', val)
        parent._attributes.append((key, val))
        attrs[key] = val

    log.debug('TRANSFORM_ATTRIBUTES [OUT]: %r' % (attrs,))
    return attrs


def transform_event(ctx, param, value):
    """Callback to transform event into title case."""
    if value is not None:
        return value.title()
    return value


def transform_resource_name(ctx, param, value):
    """Callback to transform resource_name into title case."""
    if value is not None:
        return value.title()
    return value


def process_bulk_add(ctx, param, value):
    """
    Callback to parse bulk addition of objects from a colon-delimited file.

    Format:

    + The first line of the file must be the field names.
    + Attribute pairs must be commma-separated, and in format k=v
    + The attributes must exist!
    """
    if value is None:
        return value

    # This is our object name (e.g. 'devices')
    parent_resource_name = ctx.obj.parent_resource_name
    objects = []

    # Value is already an open file handle
    reader = csv.DictReader(value, delimiter=b':')
    for row in reader:
        lineno = reader.line_num

        # Make sure the file is correctly formatted.
        if len(row) != len(reader.fieldnames):
            msg = 'File has wrong number of fields on line %d' % (lineno,)
            raise click.BadParameter(msg)

        # Transform attributes for eligible resource types
        if parent_resource_name not in NO_ATTRIBUTES:
            # FIXME(jathan): This IS going to break at some point. We need to
            # considered how to take in complex a/v pairs in this context.
            # Naively split on ',' for now.
            incoming_attributes = row['attributes'].split(',')
            outgoing_attributes = transform_attributes(
                ctx, param, incoming_attributes
            )
            row['attributes'] = outgoing_attributes

        # Transform True, False into booleans
        log.debug('FILE ROW: %r', row)
        for key, val in row.iteritems():
            # Don't evaluate dicts
            if isinstance(val, dict):
                continue

            # Evaluate strings and if they are booleans, convert them.
            if not isinstance(val, basestring):
                msg = 'Error parsing file on line %d' % (lineno,)
                raise click.BadParameter(msg)
            if val.title() in ('True', 'False'):
                row[key] = ast.literal_eval(val)
        objects.append(row)

    log.debug('PARSED BULK DATA = %r' % (objects,))

    # Return a list of dicts
    return objects


def get_resource_by_natural_key(ctx, data, resource_name, resource=None):
    """
    Attempt to return the reource_id for an object.

    :param ctx:
        Context from the calling command

    :param data:
        Query parameters used for object lookup

    :param resource_name:
        The API resource name (for display)

    :param resource:
        The API resource client object
    """
    resource_id = None
    obj = None

    # Look up the object by natural key (e.g. cidr)
    obj = ctx.obj.get_single_object(data, resource=resource)

    # If the object was found, get its id
    if obj is not None:
        resource_id = obj['id']

    # If it's not found, error out.
    if resource_id is None:
        raise click.UsageError(
            'No matching %s found; try lookup using option "-i" / "--id".' %
            (resource_name,)
        )

    return resource_id


def list_subcommand(ctx, display_fields=None, my_name=None, grep_name=None,
                    with_parent=True, return_results=False):
    """
    Determine params and a resource object to pass to ``ctx.obj.list()``

    This is used for mapping sub-commands to nested API resources.

    For example::

        nsot networks list -s 1 -c 10.0.0.0/8 subnets

    Would be mapped to::

        GET /api/sites/1/networks/5/subnets/

    If ``return_results`` is set, ``data`` will be to ``ctx.obj.detail()``
    and the results will be directly returned instead.

    :param ctx:
        Context from the calling command

    :param display_fields:
        Display fields used to list object results.

    :param my_name:
        (Optional) Overload the App's resource_name to match this value

    :param grep_name:
        (Optional) Overload the App's grep_name to match this value

    :param with_parent:
        Whether to treat the nested lookup as a detail view (default) using the
        parent resource id, or as a list view without an id lookup.

    :param return_results:
        Whether to pass the results to ``list()`` or pass them to ``detail()``
        and return them.
    """
    if display_fields is None and not return_results:
        raise SyntaxError(
            'Display fields must be provided when not returning results.'
        )

    # Gather our args from our parent and ourself
    data = ctx.parent.params
    data.update(ctx.params)

    # This will be used for resource lookup for detail routes.
    parent_resource_id = data.pop('id')

    # Prepare the app and rebase the API to include site_id.
    app = ctx.obj
    app.rebase(data)

    # Use our name, parent's command name, and the API object to retrieve the
    # endpoint resource used to call this endpoint.
    parent_resource_name = app.parent_resource_name  # e.g. 'networks'

    # If we've provided my_name, overload the App's resource_name to match
    # it.
    if my_name is None:
        my_name = ctx.info_name  # e.g. 'supernets'

    # e.g. /api/sites/1/networks/
    parent_resource = getattr(app.api, parent_resource_name)

    # Make sure that parent_resource_id is provided. This seems complicated
    # because we want to maintain dynamism across resource types.
    if with_parent:
        if parent_resource_id is None:
            if data.get('query'):
                # Enforce that we get a single item back from the query, the
                # server will return a non-2xx response on != 1 items returned,
                # which we surface as an error to the user
                data['unique'] = True

                parent_resource_id = ctx.obj.set_query(
                        data, resource=parent_resource)[0]['id']
            else:
                parent_resource_id = get_resource_by_natural_key(
                    ctx, data, parent_resource_name, parent_resource
                )

        # e.g. /api/sites/1/networks/5/supernets/
        my_resource = getattr(parent_resource(parent_resource_id), my_name)

    # Otherwise, just treat this as a list endpoint with no parent lookup.
    else:
        my_resource = getattr(parent_resource, my_name)

    app.resource_name = my_name
    app.grep_name = grep_name or my_name

    # If return_results, just pass data to my_resource.get() and return
    # whatever comes back.
    if return_results:
        return app.detail(data, resource=my_resource)

    app.list(data, display_fields=display_fields, resource=my_resource)



# -*- coding: utf-8 -*-

"""
Custom Click parameter types.
"""

from __future__ import unicode_literals

from ..vendor import click
from ..util import validate_cidr


class NetworkIdParamType(click.ParamType):
    """Custom paramer type that supports Network ID or CIDR."""
    name = 'network identifier'

    def convert(self, value, param, ctx):
        if value is None:
            return

        tests = [int, validate_cidr]
        win = False
        for test in tests:
            try:
                win = test(value)
            except:
                pass
            else:
                if not win:
                    continue
                return value

        else:
            self.fail('%s is not an valid integer or CIDR' % value, param, ctx)

    def __repr__(self):
        return 'NETWORK_ID'

NETWORK_ID = NetworkIdParamType()



# -*- coding: utf-8 -*-

"""
Sub-command for Devices.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

from __future__ import unicode_literals

from ..vendor import click
from . import callbacks


# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('hostname', 'Hostname'),
    # ('site_id': 'Site ID'),
    ('attributes', 'Attributes'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Device objects.

    A device represents various hardware components on your network such as
    routers, switches, console servers, PDUs, servers, etc.

    Devices also support arbitrary attributes similar to Networks.
    """


# Add
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Device (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-b',
    '--bulk-add',
    metavar='FILENAME',
    help='Bulk add Devices from the specified colon-delimited file.',
    type=click.File('rb'),
    callback=callbacks.process_bulk_add,
)
@click.option(
    '-H',
    '--hostname',
    metavar='HOSTNAME',
    help='The hostname of the Device.  [required]',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def add(ctx, attributes, bulk_add, hostname, site_id):
    """
    Add a new Device.

    You must provide a Site ID using the -s/--site-id option.

    When adding a new Device, you must provide a value for the -H/--hostname
    option.

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.
    """
    data = bulk_add or ctx.params

    # Enforce required options
    if bulk_add is None:
        if hostname is None:
            raise click.UsageError('Missing option "-H" / "--hostname".')

    ctx.obj.add(data)


# List
# @cli.command()
@cli.group(invoke_without_command=True)
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='Filter Devices by matching attributes (format: key=value).',
    multiple=True,
)
@click.option(
    '-d',
    '--delimited',
    is_flag=True,
    help='Display set query results separated by commas vs. newlines.',
    default=False,
    show_default=True,
)
@click.option(
    '-g',
    '--grep',
    is_flag=True,
    help='Display list results in a grep-friendly format.',
    default=False,
    show_default=True,
)
@click.option(
    '-H',
    '--hostname',
    metavar='HOSTNAME',
    help='Filter by hostname of the Device.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Device being retrieved.',
)
@click.option(
    '-N',
    '--natural-key',
    is_flag=True,
    help='Display list results by their natural key',
    default=False,
    show_default=True,
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    type=int,
    help='Limit result to N resources.',
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-q',
    '--query',
    metavar='QUERY',
    help='Perform a set query using Attributes and output matching hostnames.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def list(ctx, attributes, delimited, grep, hostname, id, limit, natural_key,
         offset, query, site_id):
    """
    List existing Devices for a Site.

    You must provide a Site ID using the -s/--site-id option.

    When listing Devices, all objects are displayed by default. You may
    optionally lookup a single Device by ID using the -i/--id option.

    You may limit the number of results using the -l/--limit option.
    """
    data = ctx.params
    data.pop('delimited')  # We don't want this going to the server.

    if ctx.invoked_subcommand is None:
        if query is not None:
            ctx.obj.natural_keys_by_query(data, delimited)
        else:
            ctx.obj.list(data, display_fields=DISPLAY_FIELDS)


@list.command()
@click.pass_context
def interfaces(ctx, *args, **kwargs):
    """Get interfaces for a Device."""
    from .cmd_interfaces import DISPLAY_FIELDS as INTERFACE_DISPLAY_FIELDS
    callbacks.list_subcommand(
        ctx, display_fields=INTERFACE_DISPLAY_FIELDS, my_name='interfaces'
    )


# Remove
@cli.command()
@click.option(
    '-H',
    '--hostname',
    'id',
    metavar='HOSTNAME',
    type=str,
    help='The hostname of the Device being deleted.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Device being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove a Device.

    You must provide a Site ID using the -s/--site-id option.

    When removing a Device, you must either provide the unique ID using
    -i/--id, or the hostname of the Device using -H/--hostname.

    If both are provided, -H/--hostname will be ignored.
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this network (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-H',
    '--hostname',
    metavar='HOSTNAME',
    help='The hostname of the Device.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Device being updated.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '--add-attributes',
    'attr_action',
    flag_value='add',
    default=True,
    help=(
        'Causes attributes to be added. This is the default and providing it '
        'will have no effect.'
    )
)
@click.option(
    '--delete-attributes',
    'attr_action',
    flag_value='delete',
    help=(
        'Causes attributes to be deleted instead of updated. If combined with'
        'with --multi the attribute will be deleted if either no value is '
        'provided, or if attribute no longer has an valid values.'
    ),
)
@click.option(
    '--replace-attributes',
    'attr_action',
    flag_value='replace',
    help=(
        'Causes attributes to be replaced instead of updated. If combined '
        ' with --multi, the entire list will be replaced.'
    ),
)
@click.option(
    '--multi',
    is_flag=True,
    help='Treat the specified attributes as a list type.',
)
@click.pass_context
def update(ctx, attributes, hostname, id, site_id, attr_action, multi):
    """
    Update a Device.

    You must provide a Site ID using the -s/--site-id option.

    When updating a Device you must provide either the unique ID (-i/--id) or
    hostname (-H/--hostname) and at least one of the optional arguments. If
    -i/--id is provided -H/--hostname will be ignored.

    If you desire to update the hostname field, you must provide -i/--id to
    uniquely identify the Device.

    The -a/--attributes option may be provided multiple times, once for each
    key-value pair.

    When modifying attributes you have three actions to choose from:

    * Add (--add-attributes). This is the default behavior that will add
    attributes if they don't exist, or update them if they do.

    * Delete (--delete-attributes). This will cause attributes to be
    deleted. If combined with --multi the attribute will be deleted if
    either no value is provided, or if the attribute no longer contains a
    valid value.

    * Replace (--replace-attributes). This will cause attributes to
    replaced. If combined with --multi and multiple attributes of the same
    name are provided, only the last value provided will be used.
    """
    if not any([attributes, hostname]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)

    if not id and not hostname:
        raise click.UsageError(
            'You must provide -H/--hostname when not providing -i/--id.'
        )

    data = ctx.params
    ctx.obj.update(data)



# -*- coding: utf-8 -*-

"""
Utilities and stuff.
"""

from __future__ import unicode_literals

from .vendor import netaddr


def get_result(response):
    """
    Get the desired result from an API response.

    :param response: Requests API response object
    """
    try:
        payload = response.json()
    except AttributeError:
        payload = response

    if 'results' in payload:
        return payload['results']

    # Or just return the payload... (next-gen)
    return payload


def validate_cidr(cidr):
    """
    Return whether ``cidr`` is valid.

    :param cidr:
        IPv4/IPv6 address
    """
    try:
        netaddr.IPNetwork(cidr)
    except (TypeError, netaddr.AddrFormatError):
        return False
    else:
        return True


def dict_to_cidr(obj):
    """
    Take an dict of a Network object and return a cidr-formatted string.

    :param obj:
        Dict of an Network object
    """
    return '%s/%s' % (obj['network_address'], obj['prefix_length'])


def slugify(s):
    """
    Slugify a string for use in URLs. This mirrors ``nsot.util.slugify()``.

    :param s:
        String to slugify
    """

    disallowed_chars = ['/']
    replacement = '_'

    for char in disallowed_chars:
        s = s.replace(char, replacement)

    return s



#!/usr/bin/env python
from __future__ import unicode_literals

"""
Generate fixtures for NSoT dev/testing.
"""

import netaddr

# from . import util as fx
import util as fx
from pynsot import client


# API_URL = 'http://localhost:8990/api'
API_URL = 'http://localhost:8000/api'
# API_URL = 'http://localhost:80/api'
EMAIL = 'admin@localhost'

api = client.EmailHeaderClient(API_URL, email=EMAIL)
site_obj = api.sites.post({'name': 'Test Site'})['data']['site']
site = api.sites(site_obj['id'])


##############
# Attributes #
##############

# Populate attributes
resource_names = ('Device', 'Network', 'Interface')
attributes = []
for resource_name in resource_names:
    attrs = fx.enumerate_attributes(resource_name)
    attributes.extend(attrs)

# Create Attribute objects
site.attributes.post(attributes)
print 'Populated Attributes.'


############
# Networks #
############

# What's our supernet to derive from?
supernet = netaddr.IPNetwork('10.16.32.0/20')  # 16x /24
# supernet = netaddr.IPNetwork('10.16.32.0/23')  # 2x /24

# Populate the /20
parents = fx.generate_networks(ipv4list=[str(supernet)])

# Populate the /24s
networks = fx.generate_networks(ipv4list=(str(n) for n in supernet.subnet(24)))

# Populate the /32s
addresses = []
for net in supernet.subnet(24):
    hosts = (str(h) for h in net.iter_hosts())
    addresses.extend(fx.generate_networks(ipv4list=hosts))

# Create Network objects
for items in (parents, networks, addresses):
    site.networks.post(items)
print 'Populated Networks.'


###########
# Devices #
###########
# 16 hosts per /24
devices = fx.generate_devices(16 * len(networks))

# Create Device objects
site.devices.post(devices)
print 'Populated Devices.'


##############
# Interfaces #
##############
# Get list of device ids
dev_resp = site.devices.get()
device_ids = [dev['id'] for dev in dev_resp['data']['devices']]

# Get list of IP addreses and make them an iterable for take()
ip_list = [a['cidr'] for a in addresses]
ip_iter = iter(ip_list)

# 16 interfaces per host
interfaces = []
for device_id in device_ids:
    my_ips = fx.take_n(16, ip_iter)
    my_interfaces = fx.generate_interfaces(device_id, address_pool=my_ips)
    interfaces.extend(my_interfaces)

# Create Interface objects
site.interfaces.post(interfaces)
print 'Populated Interfaces.'



# -*- coding: utf-8 -*-

"""
Test the Models
"""

from __future__ import unicode_literals
import pytest

from pytest import raises

from pynsot.models import Resource, Network, Device, Interface
from pynsot.util import get_result
from .fixtures import config, client, site

__all__ = ('client', 'config', 'pytest', 'site')


def test_fail_abc(client):
    '''Test that abc is doing its job'''
    class A(Resource):
        pass

    with raises(TypeError):
        a = A(client=client, site_id=1)
        return a


def test_correct_site(client, site):
    '''Tests that a resource gets sent to the correct site_id'''
    n = Network(client=client, site_id=site['id'], cidr='8.8.8.0/24')
    assert n.ensure()
    assert n.existing_resource()['site_id'] == site['id']

    from pynsot.util import get_result
    manual = get_result(client.sites(site['id']).networks('8.8.8.0/24').get())
    assert manual['site_id'] == site['id']


def test_args(client, site):
    '''Tests for exceptions that should be thrown in certain arg combos'''
    with raises(TypeError):
        # not including cidr or raw
        Network(client=client, site=site['id'])
    with raises(TypeError):
        # including raw, but not raw['site_id']
        Network(client=client, raw={})
    with raises(TypeError):
        # not including hostname or raw
        Device(client=client, site=site['id'])
    with raises(TypeError):
        # not including name+device or raw
        Interface(client=client, site=site['id'])


def test_raw_precedence(client, site):
    '''Makes sure raw kwarg site_id takes precedence'''
    # first create a resource to get raw from
    n = Network(client=client, site_id=site['id'], cidr='8.8.8.0/24')
    assert n.ensure()

    # Make sure that raw settings take precedence
    raw = n.existing_resource()
    n2 = Network(client=client, site=999, raw=raw)
    assert n2['site_id'] != 999


def test_existing(client, site):
    '''Test functionality around existing resource and caching of the result'''
    site_id = site['id']
    c = client

    n = Network(client=c, site_id=site_id, cidr='8.8.8.0/24')
    assert n.purge()
    assert not n.exists()
    assert n._existing_resource == {}
    assert n.existing_resource() == n._existing_resource
    assert n.ensure()
    # Make sure cache clearing works
    assert n._existing_resource == {}
    assert n.exists()
    assert n.existing_resource() == n._existing_resource


def test_net_closest_parent(client, site):
    '''Test that Network.closest_parent returns instance of Network or dict'''
    site_id = site['id']
    c = client

    parent = Network(client=c, site_id=site_id, cidr='8.8.8.0/24')
    assert parent.ensure()

    child = Network(client=c, site_id=site_id, cidr='8.8.8.8/32')
    assert child.closest_parent() == parent

    orphan = Network(client=c, site_id=site_id, cidr='1.1.1.1/32')
    assert not orphan.closest_parent()
    assert orphan.closest_parent() == {}


def test_dict():
    '''Test methods/behavior that should work like a dictionary'''
    n = Network(site_id=1, cidr='8.8.8.0/24')
    assert n['site_id'] == 1
    n['site_id'] = 2
    assert n['site_id'] == 2

    assert n.keys()
    assert n.items()
    assert dict(n)


def test_payload_not_none_raw_and_not(client, site):
    '''Make sure payload gets set fine for both EZ and raw approaches

    Also make sure both instances are the same, using comparisons
    '''
    n = Network(client=client, site_id=site['id'], cidr='8.8.8.0/24')
    assert n.payload
    assert n.ensure()

    n2 = Network(
        raw=get_result(
            client.sites(site['id']).networks('8.8.8.0/24').get()
        )
    )

    assert n2.payload

    # Test some magic methods on raw-init'd instance
    assert len(n2)
    assert n == n2
    assert n2.keys()
    assert n2.items()


def test_clear_cache_on_change(client, site):
    '''Test that cache is cleared on any change to the instance'''
    site_id = site['id']
    c = client

    n = Network(client=c, site_id=site_id, cidr='8.8.8.0/24')
    assert n.ensure()
    assert n.exists()

    non_existing_site = get_result(c.sites.get())[-1]['id'] + 1000
    n['site_id'] = non_existing_site
    # First assert that the cache property was cleared
    assert not n._existing_resource
    assert not n.existing_resource()
    # assert that the resource isn't successfully looked up if site doesn't
    # match
    assert not n.exists()
    # Change site back and test
    n['site_id'] = site_id
    assert n.exists()


def test_ip4_host():
    '''Test to make sure IPv4 host works fine'''
    net = '8.8.8.8/32'
    n = Network(site_id=1, cidr=net)
    assert n['network_address'] == '8.8.8.8'
    assert n['prefix_length'] == 32
    assert n.identifier == net
    assert n['site_id'] == 1
    assert n.is_host
    assert n.resource_name == 'networks'
    assert n['state'] == 'assigned'
    assert dict(n)
    assert n['attributes'] == {}
    assert n['network_address']


def test_ip4_net(client):
    '''Test to make sure IPv4 subnet works fine'''
    net = '8.8.8.0/24'
    n = Network(client=client, site_id=1, cidr=net)
    assert n['network_address'] == '8.8.8.0'
    assert n['prefix_length'] == 24
    assert n.identifier == net
    assert n['site_id'] == 1
    assert not n.is_host
    assert n.resource_name == 'networks'
    assert n['state'] == 'allocated'
    assert dict(n)
    assert n['network_address']


def test_ip6_net(client):
    '''Test to make sure IPv6 subnet works fine'''
    net = '2001::/64'
    n = Network(client=client, site_id=1, cidr=net)
    assert n['network_address'] == '2001::'
    assert n['prefix_length'] == 64
    assert n.identifier == net
    assert n['site_id'] == 1
    assert not n.is_host
    assert n.resource_name == 'networks'
    assert n['state'] == 'allocated'
    assert dict(n)
    assert n['network_address']


def test_ip6_host(client):
    '''Test to make sure IPv6 host works fine'''
    net = '2001::1/128'
    n = Network(client=client, site_id=1, cidr=net)
    assert n['network_address'] == '2001::1'
    assert n['prefix_length'] == 128
    assert n.identifier == net
    assert n['site_id'] == 1
    assert n.is_host
    assert n.resource_name == 'networks'
    assert n['state'] == 'assigned'
    assert dict(n)
    assert n['network_address']


def test_device(client):
    '''Test to make sure device works fine'''
    name = 'pytest'
    d = Device(client=client, site_id=1, hostname=name)
    assert d['hostname'] == name
    assert d.identifier == name
    assert d['site_id'] == 1
    assert d.resource_name == 'devices'
    assert dict(d)
    assert d['hostname']
    assert d['attributes'] == {}
    d['attributes'] = {'desc': 'test host'}


def test_device_eq(client):
    '''Test to make sure device comparison works fine'''
    name = 'pytest'
    d1 = Device(client=client, site_id=1, hostname=name)
    d2 = Device(client=client, site_id=1, hostname=name)
    assert d1 == d2


def test_interface(client):
    '''Test to make sure interface init works fine'''
    name = 'eth0'
    i = Interface(client=client, site_id=1, name=name, device=1)
    assert i['name'] == name
    assert i['device'] == 1
    assert i['site_id'] == 1
    assert i.resource_name == 'interfaces'
    assert dict(i)
    assert i['name']
    assert i['attributes'] == {}
    i['attributes'] = {'desc': 'test host'}


def test_interface_eq(client):
    '''Test to make sure interface comparison works fine'''
    name = 'eth0'
    i1 = Interface(client=client, site_id=1, name=name, device=1)
    i2 = Interface(client=client, site_id=1, name=name, device=1)
    assert i1 == i2


def test_ip4_send(client, site):
    '''Test upstream write actions for IPv4'''

    site_id = site['id']
    subnet = Network(client=client, site_id=site_id, cidr='254.0.0.0/24')
    host = Network(client=client, site_id=site_id, cidr='254.0.0.1/32')

    assert subnet.purge()
    assert not subnet.exists()
    assert subnet.ensure()
    assert subnet.exists()
    assert host.purge()
    assert not host.exists()
    assert host.ensure()
    assert host.exists()

    host.purge()
    subnet.purge()
    assert not all([subnet.exists(), host.exists()])


def test_device_send(client, site):
    '''Test upstream write actions for devices'''
    site_id = site['id']
    name = 'pytest'
    d = Device(client=client, site_id=site_id, hostname=name)
    assert d.purge()
    assert not d.exists()
    assert d.ensure()
    assert d.exists()
    assert d.purge()
    assert not d.exists()


# This section is commented out because some changes need make to interfaces to
# better support friendly-specifying hostname vs. device ID during
# instantiation.
#
# def test_interface_send(client, site):
#    '''Test upstream write actions for interfaces'''
#     site_id = site['id']
#     host = 'intftest'
#     d = Device(client=client, site_id=site_id, hostname=host)
#     assert d.ensure()
#
#     i = Interface(client=client, site_id=site_id, name='eth0', device=host)
#
#     assert i.purge()
#     assert not i.exists()
#     assert i.ensure()
#     assert i.exists()
#     assert i.purge()
#     assert not i.exists()
#
#     assert d.purge()
#     assert not d.exists()
#
#
# def test_interface_wo_device(client, site):
#    '''Test to make sure device is set to 0 if doesnt exist'''
#     site_id = site['id']
#     host = 'doesnt-exist-yet'
#     d = Device(client=client, site_id=site_id, hostname=host)
#     assert d.purge()
#
#     i = Interface(client=client, site_id=site_id, name='eth0', device=host)
#     assert i
#     assert i['device'] == 0
#
#     assert d.ensure()
#
#     # The device id is checked every time __iter__ is called
#     assert i['device'] != 0
#
#     d.purge()
